class ManagerError(Exception):
    """Base class for terminal manager errors."""

    def __init__(self, message: str, exc: Exception | None = None):
        super().__init__()
        self.exc = exc
        self.message = message
        self.details = str(exc) if exc else None

    def __str__(self):
        return f"{self.message} ({self.details})" if self.details else self.message


class CommandError(ManagerError):
    """Error to indicate that the command execution failed."""


class NameKeyError(ManagerError):
    """Error to indicate that name and key are not defined."""

    def __init__(self) -> None:
        super().__init__("Name and key not defined")


class InvalidSensorError(ManagerError):
    """Error to indicate that a sensor is invalid."""

    def __init__(self, key: str, details: str) -> None:
        super().__init__(f"Sensor '{key}' is invalid")
        self.details = details
        self.key = key


class CommandLoopError(Exception):
    """Error to indicate that a command loop was detected."""

    def __init__(self, key: str) -> None:
        super().__init__("Command loop detected")
        self.details = f"Key: '{key}'"
        self.key = key



"""Terminal manager."""

from __future__ import annotations

from collections.abc import Sequence
from contextlib import suppress
from dataclasses import dataclass
import logging
from typing import Any

from .collection import Collection
from .command import PLACEHOLDER_KEY, ActionCommand, Command, SensorCommand
from .default_collections import ActionKey, SensorKey
from .errors import (
    CommandError,
    CommandLoopError,
    InvalidSensorError,
    ManagerError,
    NameKeyError,
)
from .event import Event
from .sensor import BinarySensor, NumberSensor, Sensor, TextSensor, VersionSensor
from .synchronizer import Synchronizer

_LOGGER = logging.getLogger(__name__)

DEFAULT_NAME = "Manager"
DEFAULT_COMMAND_TIMEOUT = 15
DEFAULT_ALLOW_TURN_OFF = False


@dataclass(frozen=True)
class CommandOutput:
    command_string: str
    timestamp: float
    stdout: list[str]
    stderr: list[str]
    code: int


class Manager(Collection, Synchronizer):
    def __init__(
        self,
        *,
        name: str = DEFAULT_NAME,
        command_timeout: int = DEFAULT_COMMAND_TIMEOUT,
        allow_turn_off: bool = DEFAULT_ALLOW_TURN_OFF,
        collection: Collection | None = None,
        logger: logging.Logger = _LOGGER,
    ) -> None:
        Synchronizer.__init__(self)
        Collection.__init__(
            self,
            name,
            collection.action_commands if collection else None,
            collection.sensor_commands if collection else None,
        )
        self.command_timeout = command_timeout
        self.allow_turn_off = allow_turn_off
        self.logger = logger

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.async_close()

    @property
    def network_interface(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.NETWORK_INTERFACE)

    @property
    def mac_address(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.MAC_ADDRESS)

    @property
    def wake_on_lan(self) -> bool | None:
        return self._last_known_value_or_none(SensorKey.WAKE_ON_LAN)

    @property
    def machine_type(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.MACHINE_TYPE)

    @property
    def hostname(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.HOSTNAME)

    @property
    def os_name(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.OS_NAME)

    @property
    def os_version(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.OS_VERSION)

    @property
    def os_architecture(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.OS_ARCHITECTURE)

    @property
    def device_name(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.DEVICE_NAME)

    @property
    def device_model(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.DEVICE_MODEL)

    @property
    def manufacturer(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.MANUFACTURER)

    @property
    def serial_number(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.SERIAL_NUMBER)

    @property
    def cpu_name(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.CPU_NAME)

    @property
    def cpu_cores(self) -> int | None:
        return self._last_known_value_or_none(SensorKey.CPU_CORES)

    @property
    def cpu_hardware(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.CPU_HARDWARE)

    @property
    def cpu_model(self) -> str | None:
        return self._last_known_value_or_none(SensorKey.CPU_MODEL)

    def _last_known_value_or_none(self, sensor_key: str) -> Any | None:
        if sensor := self.sensors_by_key.get(sensor_key):
            return sensor.last_known_value

        return None

    def _clear_sensors(self) -> None:
        for command in self.sensor_commands:
            command.update_sensors(self, None)

    async def async_close(self) -> None:
        """Close."""

    async def async_update_sensor_commands(self, force: bool = False) -> None:
        """Update the sensor commands.

        Execute sensor commands that passed their `interval` or
        all sensor commands with `force=True`.
        """
        for command in self.sensor_commands:
            if not (force or command.should_update):
                return
            with suppress(CommandError):
                await self.async_execute_command(command)

    async def async_execute_command_string(
        self,
        string: str,
        command_timeout: int | None = None,
    ) -> CommandOutput:
        """Execute a command string.

        Raises:
            CommandError

        """
        raise CommandError("Not implemented")

    async def async_execute_command(
        self,
        command: Command,
        variables: dict | None = None,
    ) -> CommandOutput:
        """Execute a command.

        Raises:
            CommandError

        """
        try:
            output = await command.async_execute(self, variables)
        except CommandError as exc:
            self.logger.debug("%s: %s => %s", self.name, command.string, exc)
            raise

        self.logger.debug(
            "%s: %s => %s, %s, %s",
            self.name,
            output.command_string,
            output.stdout,
            output.stderr,
            output.code,
        )

        return output

    async def async_run_action(
        self,
        key: str,
        variables: dict | None = None,
    ) -> CommandOutput:
        """Run an action.

        Raises:
            CommandError

        """
        command = self.get_action_command(key)
        return await self.async_execute_command(command, variables)

    async def async_poll_sensor(
        self,
        key: str,
        *,
        raise_errors: bool = False,
    ) -> Sensor:
        """Poll a sensor.

        Raises:
            CommandError (only with `raise_errors=True`)

        """
        sensors = await self.async_poll_sensors([key], raise_errors=raise_errors)
        return sensors[0]

    async def async_poll_sensors(
        self,
        keys: Sequence[str],
        *,
        raise_errors: bool = False,
    ) -> list[Sensor]:
        """Poll multiple sensors.

        Raises:
            CommandError (only with `raise_errors=True`)

        """
        sensors = [self.get_sensor(key) for key in keys]
        commands = [self.get_sensor_command(key) for key in set(keys)]

        for command in commands:
            try:
                await self.async_execute_command(command)
            except CommandError:
                if raise_errors:
                    raise

        return sensors

    async def async_set_sensor_value(
        self,
        key: str,
        value: Any,
        *,
        raise_errors: bool = False,
    ) -> Sensor:
        """Set the value of a controllable sensor.

        Raises:
            TypeError (only with `raise_errors=True`)
            ValueError (only with `raise_errors=True`)
            CommandError (only with `raise_errors=True`)

        """
        sensors = await self.async_set_sensor_values(
            [key], [value], raise_errors=raise_errors
        )
        return sensors[0]

    async def async_set_sensor_values(
        self,
        keys: Sequence[str],
        values: Sequence[Any],
        *,
        raise_errors: bool = False,
    ) -> list[Sensor]:
        """Set the value of multiple controllable sensors.

        Raises:
            TypeError (only with `raise_errors=True`)
            ValueError (only with `raise_errors=True`)
            CommandError (only with `raise_errors=True`)

        """
        sensors = await self.async_poll_sensors(keys, raise_errors=raise_errors)

        for i, sensor in enumerate(sensors):
            if sensor.value is None:
                continue
            if callable(values[i]):
                value = values[i](sensor.value)
            else:
                value = values[i]
            try:
                await sensor.async_set(self, value)
            except (TypeError, ValueError, CommandError):
                if raise_errors:
                    raise

        return await self.async_poll_sensors(keys, raise_errors=raise_errors)

    async def async_turn_off(self) -> CommandOutput:
        """Turn off by running the `TURN_OFF` action.

        Raises:
            PermissionError
            CommandError

        """
        if not self.allow_turn_off:
            raise PermissionError("Not allowed to turn off")

        return await self.async_run_action(ActionKey.TURN_OFF)

    async def async_restart(self) -> CommandOutput:
        """Restart by running the `RESTART` action.

        Raises:
            CommandError

        """
        return await self.async_run_action(ActionKey.RESTART)



"""The Windows cmd collection."""

from terminal_manager.collection import Collection
from terminal_manager.command import ActionCommand, SensorCommand
from terminal_manager.sensor import NumberSensor, TextSensor

from .const import ActionKey, ActionName, SensorKey, SensorName

windows_cmd = Collection(
    "Windows",
    [
        ActionCommand(
            "shutdown /s /t 0",
            ActionName.TURN_OFF,
            ActionKey.TURN_OFF,
        ),
        ActionCommand(
            "shutdown /r /t 0",
            ActionName.RESTART,
            ActionKey.RESTART,
        ),
    ],
    [
        # TODO: WAKE_ON_LAN
        SensorCommand(
            "for /f %i in ('wmic path win32_ip4routetable "
            "where \"Destination='0.0.0.0'\" "
            "get InterfaceIndex ^| "
            'findstr /r "\\<[0-9][0-9]*\\>"\') do '
            '@for /f "skip=2 tokens=2,3 delims=," %j in (\'wmic nic '
            'where "InterfaceIndex=%i" '
            "get MACAddress^,NetConnectionID /format:csv') do "
            "@echo %j & @echo %k",
            sensors=[
                TextSensor(
                    SensorName.MAC_ADDRESS,
                    SensorKey.MAC_ADDRESS,
                ),
                TextSensor(
                    SensorName.NETWORK_INTERFACE,
                    SensorKey.NETWORK_INTERFACE,
                ),
            ],
        ),
        SensorCommand(
            "hostname",
            sensors=[
                TextSensor(
                    SensorName.HOSTNAME,
                    SensorKey.HOSTNAME,
                )
            ],
        ),
        SensorCommand(
            'for /f "skip=1 tokens=*" %i in (\'wmic ComputerSystem '
            "get SystemType') do "
            "@echo %i",
            sensors=[
                TextSensor(
                    SensorName.MACHINE_TYPE,
                    SensorKey.MACHINE_TYPE,
                ),
            ],
        ),
        SensorCommand(
            "for /f \"skip=1 tokens=*\" %i in ('wmic OS get Caption') do @echo %i",
            sensors=[
                TextSensor(
                    SensorName.OS_NAME,
                    SensorKey.OS_NAME,
                ),
            ],
        ),
        SensorCommand(
            "for /f \"skip=1\" %i in ('wmic OS get Version') do @echo %i",
            sensors=[
                TextSensor(
                    SensorName.OS_VERSION,
                    SensorKey.OS_VERSION,
                ),
            ],
        ),
        SensorCommand(
            "for /f \"skip=1\" %i in ('wmic OS get OSArchitecture') do @echo %i",
            sensors=[
                TextSensor(
                    SensorName.OS_ARCHITECTURE,
                    SensorKey.OS_ARCHITECTURE,
                ),
            ],
        ),
        SensorCommand(
            "for /f \"skip=1 tokens=*\" %i in ('wmic csproduct get Name') do @echo %i",
            sensors=[
                TextSensor(
                    SensorName.DEVICE_NAME,
                    SensorKey.DEVICE_NAME,
                ),
            ],
        ),
        SensorCommand(
            "for /f \"skip=1 tokens=*\" %i in ('wmic csproduct get Version') do @echo %i",
            sensors=[
                TextSensor(
                    SensorName.DEVICE_MODEL,
                    SensorKey.DEVICE_MODEL,
                ),
            ],
        ),
        SensorCommand(
            "for /f \"skip=1 tokens=*\" %i in ('wmic csproduct get Vendor') do @echo %i",
            sensors=[
                TextSensor(
                    SensorName.MANUFACTURER,
                    SensorKey.MANUFACTURER,
                ),
            ],
        ),
        SensorCommand(
            "for /f \"skip=1 tokens=*\" %i in ('wmic bios get SerialNumber') do @echo %i",
            sensors=[
                TextSensor(
                    SensorName.SERIAL_NUMBER,
                    SensorKey.SERIAL_NUMBER,
                ),
            ],
        ),
        SensorCommand(
            "for /f \"skip=1 tokens=*\" %i in ('wmic cpu get Name') do @echo %i",
            sensors=[
                TextSensor(
                    SensorName.CPU_NAME,
                    SensorKey.CPU_NAME,
                ),
            ],
        ),
        SensorCommand(
            "for /f \"skip=1\" %i in ('wmic cpu get NumberOfCores') do @echo %i",
            sensors=[
                NumberSensor(
                    SensorName.CPU_CORES,
                    SensorKey.CPU_CORES,
                ),
            ],
        ),
        SensorCommand(
            'for /f "skip=1" %i in (\'wmic ComputerSystem '
            "get TotalPhysicalMemory') do "
            "@echo %i",
            sensors=[
                NumberSensor(
                    SensorName.TOTAL_MEMORY,
                    SensorKey.TOTAL_MEMORY,
                    unit="B",
                ),
            ],
        ),
        SensorCommand(
            'for /f "skip=1" %i in (\'wmic OS '
            "get FreePhysicalMemory') do "
            "@echo %i",
            interval=30,
            sensors=[
                NumberSensor(
                    SensorName.FREE_MEMORY,
                    SensorKey.FREE_MEMORY,
                    unit="kB",
                ),
            ],
        ),
        SensorCommand(
            'for /f "tokens=1,2" %i in (\'wmic LogicalDisk '
            "get DeviceID^,FreeSpace ^| "
            'findstr ":"\') do '
            "@echo %i^|%j",
            interval=60,
            separator="|",
            sensors=[
                NumberSensor(
                    SensorName.FREE_DISK_SPACE,
                    SensorKey.FREE_DISK_SPACE,
                    dynamic=True,
                    unit="B",
                )
            ],
        ),
        SensorCommand(
            "for /f \"skip=1\" %i in ('wmic CPU get LoadPercentage') do @echo %i",
            interval=30,
            sensors=[
                NumberSensor(
                    SensorName.CPU_LOAD,
                    SensorKey.CPU_LOAD,
                    unit="%",
                ),
            ],
        ),
        SensorCommand(
            "for /f %i in ('wmic /namespace:\\\\root\\wmi "
            "path MSAcpi_ThermalZoneTemperature "
            "get CurrentTemperature ^| "
            'findstr /r "\\<[0-9][0-9]*\\>"\') do '
            "@set /a x=(%i - 2732) / 10",
            interval=60,
            sensors=[
                NumberSensor(
                    SensorName.TEMPERATURE,
                    SensorKey.TEMPERATURE,
                    unit="°C",
                )
            ],
        ),
        SensorCommand(
            "wmic process get processId | "
            'findstr /r "\\<[0-9][0-9]*\\>" | '
            'find /c /v ""',
            interval=60,
            sensors=[
                NumberSensor(
                    SensorName.PROCESSES,
                    SensorKey.PROCESSES,
                )
            ],
        ),
    ],
)



"""The Windows PowerShell collection."""

from terminal_manager.collection import Collection
from terminal_manager.command import ActionCommand, SensorCommand
from terminal_manager.sensor import BinarySensor, NumberSensor, TextSensor

from .const import ActionKey, ActionName, SensorKey, SensorName

windows_ps = Collection(
    "Windows (Power Shell)",
    [
        ActionCommand(
            "Stop-Computer -Force",
            ActionName.TURN_OFF,
            ActionKey.TURN_OFF,
        ),
        ActionCommand(
            "Restart-Computer -Force",
            ActionName.RESTART,
            ActionKey.RESTART,
        ),
    ],
    [
        SensorCommand(
            "$x = Get-NetAdapterPowerManagement "
            '-Name "&{network_interface}" | '
            "Select WakeOnMagicPacket; "
            "$x.WakeOnMagicPacket",
            sensors=[
                BinarySensor(
                    SensorName.WAKE_ON_LAN,
                    SensorKey.WAKE_ON_LAN,
                )
            ],
        ),
        SensorCommand(
            "$y = Get-CimInstance Win32_IP4RouteTable "
            "-Filter \"Destination='0.0.0.0'\" | "
            "Select InterfaceIndex; "
            "$x = Get-CimInstance Win32_NetworkAdapter "
            "-Property NetConnectionID, InterfaceIndex, MACAddress "
            '-Filter "InterfaceIndex=$($y.InterfaceIndex)" | '
            "Select MACAddress,NetConnectionID; "
            "$x.MACAddress; "
            "$x.NetConnectionID",
            sensors=[
                TextSensor(
                    SensorName.MAC_ADDRESS,
                    SensorKey.MAC_ADDRESS,
                ),
                TextSensor(
                    SensorName.NETWORK_INTERFACE,
                    SensorKey.NETWORK_INTERFACE,
                ),
            ],
        ),
        SensorCommand(
            "$x = Get-CimInstance Win32_ComputerSystem | "
            "Select Name,SystemType; "
            "$x.Name; "
            "$x.SystemType",
            sensors=[
                TextSensor(
                    SensorName.HOSTNAME,
                    SensorKey.HOSTNAME,
                ),
                TextSensor(
                    SensorName.MACHINE_TYPE,
                    SensorKey.MACHINE_TYPE,
                ),
            ],
        ),
        SensorCommand(
            "$x = Get-CimInstance Win32_OperatingSystem | "
            "Select Caption,Version,OSArchitecture; "
            "$x.Caption; "
            "$x.Version; "
            "$x.OSArchitecture",
            sensors=[
                TextSensor(
                    SensorName.OS_NAME,
                    SensorKey.OS_NAME,
                ),
                TextSensor(
                    SensorName.OS_VERSION,
                    SensorKey.OS_VERSION,
                ),
                TextSensor(
                    SensorName.OS_ARCHITECTURE,
                    SensorKey.OS_ARCHITECTURE,
                ),
            ],
        ),
        SensorCommand(
            "$x = Get-CimInstance Win32_ComputerSystemProduct | "
            "Select Vendor,Name,Version; "
            "$x.Name; "
            "$x.Version; "
            "$x.Vendor",
            sensors=[
                TextSensor(
                    SensorName.DEVICE_NAME,
                    SensorKey.DEVICE_NAME,
                ),
                TextSensor(
                    SensorName.DEVICE_MODEL,
                    SensorKey.DEVICE_MODEL,
                ),
                TextSensor(
                    SensorName.MANUFACTURER,
                    SensorKey.MANUFACTURER,
                ),
            ],
        ),
        SensorCommand(
            "$x = Get-CimInstance Win32_BIOS | "
            "Select SerialNumber; "
            "$x.SerialNumber",
            sensors=[
                TextSensor(
                    SensorName.SERIAL_NUMBER,
                    SensorKey.SERIAL_NUMBER,
                )
            ],
        ),
        SensorCommand(
            "$x = Get-CimInstance Win32_Processor | "
            "Select Name,NumberOfCores; "
            "$x.Name; "
            "$x.NumberOfCores",
            sensors=[
                TextSensor(
                    SensorName.CPU_NAME,
                    SensorKey.CPU_NAME,
                ),
                NumberSensor(
                    SensorName.CPU_CORES,
                    SensorKey.CPU_CORES,
                ),
            ],
        ),
        SensorCommand(
            "$x = Get-CimInstance Win32_ComputerSystem | "
            "Select TotalPhysicalMemory; "
            "$x.TotalPhysicalMemory",
            sensors=[
                NumberSensor(
                    SensorName.TOTAL_MEMORY,
                    SensorKey.TOTAL_MEMORY,
                    unit="B",
                )
            ],
        ),
        SensorCommand(
            "$x = Get-CimInstance Win32_OperatingSystem | "
            "Select FreePhysicalMemory; "
            "$x.FreePhysicalMemory",
            interval=30,
            sensors=[
                NumberSensor(
                    SensorName.FREE_MEMORY,
                    SensorKey.FREE_MEMORY,
                    unit="kB",
                )
            ],
        ),
        SensorCommand(
            "Get-CimInstance Win32_LogicalDisk | "
            "Select DeviceID,FreeSpace | "
            'ForEach-Object {$_.DeviceID + "|" + $_.FreeSpace}',
            interval=60,
            separator="|",
            sensors=[
                NumberSensor(
                    SensorName.FREE_DISK_SPACE,
                    SensorKey.FREE_DISK_SPACE,
                    dynamic=True,
                    unit="B",
                )
            ],
        ),
        SensorCommand(
            "$x = Get-CimInstance Win32_Processor | "
            "Select LoadPercentage; "
            "$x.LoadPercentage",
            interval=30,
            sensors=[
                NumberSensor(
                    SensorName.CPU_LOAD,
                    SensorKey.CPU_LOAD,
                    unit="%",
                )
            ],
        ),
        SensorCommand(
            "$x = Get-CimInstance msacpi_thermalzonetemperature "
            '-namespace "root/wmi" | '
            "Select CurrentTemperature; "
            "($x.CurrentTemperature - 2732) / 10",
            interval=60,
            sensors=[
                NumberSensor(
                    SensorName.TEMPERATURE,
                    SensorKey.TEMPERATURE,
                    unit="°C",
                )
            ],
        ),
        SensorCommand(
            "Get-Process | Measure | ForEach-Object {$_.Count}",
            interval=60,
            sensors=[
                NumberSensor(
                    SensorName.PROCESSES,
                    SensorKey.PROCESSES,
                )
            ],
        ),
    ],
)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import KW_ONLY, dataclass, field
import re
from string import Template
from time import time
from typing import TYPE_CHECKING

from .errors import CommandError, CommandLoopError, InvalidSensorError
from .helpers import name_to_key
from .sensor import Sensor

if TYPE_CHECKING:
    from . import Collection, CommandOutput, Manager

SENSOR_DELIMITER = "&"
VARIABLE_DELIMITER = "@"
PLACEHOLDER_KEY = "_"


# Not needed anymore after python 3.11
class Template(Template):
    def get_identifiers(self):
        ids = []
        for mo in self.pattern.finditer(self.template):
            named = mo.group("named") or mo.group("braced")
            if named is not None and named not in ids:
                ids.append(named)
            elif (
                named is None
                and mo.group("invalid") is None
                and mo.group("escaped") is None
            ):
                raise ValueError("Unrecognized named group in pattern", self.pattern)
        return ids


class SensorTemplate(Template):
    delimiter = SENSOR_DELIMITER
    pattern = rf"""
    {re.escape(delimiter)}{{(?:
      (?P<braced>{Template.idpattern})|
      (?P<invalid>)|
      (?P<named>)|
      (?P<escaped>)
    )}}
    """


class VariableTemplate(Template):
    delimiter = VARIABLE_DELIMITER
    pattern = rf"""
    {re.escape(delimiter)}{{(?:
      (?P<braced>{Template.idpattern})|
      (?P<invalid>)|
      (?P<named>)|
      (?P<escaped>)
    )}}
    """


class DynamicData:
    def __init__(
        self,
        sensor: Sensor,
        id_field: str,
        data_field: str,
        name_field: str | None,
    ) -> None:
        self.id = id_field.strip()
        name = name_field.strip() if name_field else self.id
        self.key = f"{sensor.key}_{name_to_key(self.id)}"
        self.name = f"{sensor.name} {name}" if sensor.name else name
        self.data = data_field


@dataclass
class Command:
    string: str
    _: KW_ONLY
    timeout: float | None = None
    renderer: Callable[[str], str] | None = None
    _linked_sensors: set[str] = field(default_factory=set, init=False, repr=False)

    @property
    def required_variables(self) -> set[str]:
        """Variables required to render the command string."""
        return set(VariableTemplate(self.string).get_identifiers())

    @property
    def required_sensors(self) -> set[str]:
        """Sensors required to render the command string."""
        return set(SensorTemplate(self.string).get_identifiers())

    @property
    def linked_sensors(self) -> set[str]:
        """Sensors to poll after execution of the command."""
        return self._linked_sensors

    @property
    def sub_sensors(self) -> set[str]:
        """Set of required and linked sensors."""
        return {*self.required_sensors, *self.linked_sensors}

    def _render(self, string: str) -> str:
        if self.renderer:
            string = self.renderer(string)

        return string

    def check(self, collection: Collection) -> None:
        """Check configuration.

        Raises:
            InvalidSensorError
            CommandLoopError

        """
        commands_by_key = collection.sensor_commands_by_sensor_key
        commands = []

        def detect_loop(command: Command) -> None:
            sub_commands = []
            if command not in commands:
                commands.append(command)
            for key in command.sub_sensors:
                if key not in commands_by_key:
                    continue
                if (sub_command := commands_by_key[key]) in commands:
                    raise CommandLoopError(key)
                if sub_command not in sub_commands:
                    sub_commands.append(sub_command)
                    detect_loop(sub_command)

        detect_loop(self)

    async def async_execute(
        self,
        manager: Manager,
        variables: dict | None = None,
    ) -> CommandOutput:
        """Execute.

        Raises:
            CommandError

        """
        variables = variables or {}
        sensor_values_by_key = {}
        string = self.string

        try:
            self.check(manager)
        except (CommandLoopError, InvalidSensorError) as exc:
            raise CommandError("Command check failed", exc) from exc

        try:
            string = VariableTemplate(string).substitute(variables)
        except Exception as exc:
            raise CommandError("Failed to substitute variable", exc) from exc

        try:
            sensors = await manager.async_poll_sensors(self.required_sensors)
        except Exception as exc:
            raise CommandError("Failed to poll required sensors", exc) from exc

        for sensor in sensors:
            if sensor.value is not None:
                sensor_values_by_key[sensor.key] = sensor.value
            else:
                raise CommandError(f"Value of required sensor {sensor.key} is None")

        try:
            string = SensorTemplate(string).substitute(sensor_values_by_key)
        except Exception as exc:
            raise CommandError("Failed to substitute sensor", exc) from exc

        try:
            string = self._render(string)
        except Exception as exc:
            raise CommandError("Failed to render string", exc) from exc

        output = await manager.async_execute_command_string(string, self.timeout)

        await manager.async_poll_sensors(self.linked_sensors, raise_errors=True)

        return output


@dataclass
class ActionCommand(Command):
    name: str | None = None
    key: str | None = None
    _: KW_ONLY
    attributes: dict = field(default_factory=dict)

    def __post_init__(self):
        self.key = self.key or name_to_key(self.name)


@dataclass
class SensorCommand(Command):
    _: KW_ONLY
    interval: int | None = None
    separator: str | None = None
    sensors: list[Sensor] = field(default_factory=list)

    def __post_init__(self):
        self.last_update: float | None = None

    @property
    def linked_sensors(self) -> set[str]:
        linked_sensors = {*self._linked_sensors}

        for sensor in self.sensors:
            linked_sensors.update(sensor.linked_sensors)

        return {key for key in linked_sensors if key not in self.sensors_by_key}

    @property
    def should_update(self) -> bool:
        if not self.interval:
            return False

        if not self.last_update:
            return True

        if time() - self.last_update < self.interval:
            return False

        return True

    @property
    def sensors_by_key(self) -> dict[str, Sensor]:
        return {
            sensor.key: sensor
            for command_sensor in self.sensors
            for sensor in (command_sensor, *command_sensor.child_sensors)
            if command_sensor.key != PLACEHOLDER_KEY
        }

    def check(self, collection: Collection) -> None:
        """Check command configuration.

        Raises:
            InvalidSensorError
            CommandLoopError

        """
        dynamic = False

        for sensor in self.sensors:
            if sensor.dynamic:
                dynamic = True
            if dynamic and not sensor.dynamic:
                raise InvalidSensorError(sensor.key, "Sensor must be dynamic")
            sensor.check(collection)

        super().check(collection)

    def remove_sensor(self, key: str) -> None:
        """Remove a sensor."""
        self.sensors = [
            Sensor(key=PLACEHOLDER_KEY) if sensor.key == key else sensor
            for sensor in self.sensors
        ]

    def update_sensors(self, manager: Manager, output: CommandOutput | None) -> None:
        """Update the sensors."""
        if output and output.code == 0:
            self.last_update = output.timestamp
            data = output.stdout
        else:
            data = []

        dyn_start = None

        for i, sensor in enumerate(self.sensors):
            if sensor.dynamic:
                dyn_start = i
                break
            sensor_data = data[i] if len(data) > i else None
            sensor.update(manager, sensor_data)

        if dyn_start is None:
            return

        dyn_data = data[dyn_start:]
        dyn_sensors = self.sensors[dyn_start:]
        dyn_count = len(dyn_sensors)

        for i, sensor in enumerate(dyn_sensors):
            sensor_data = [
                DynamicData(
                    sensor,
                    fields[0],
                    fields[i + 1],
                    fields[dyn_count + 1] if len(fields) > dyn_count + 1 else None,
                )
                for line in dyn_data
                if len(fields := line.split(self.separator)) > dyn_count
            ] or None
            sensor.update(manager, sensor_data or None)

    async def async_execute(
        self,
        manager: Manager,
        variables: dict | None = None,
    ) -> CommandOutput:
        try:
            output = await super().async_execute(manager, variables)
        except CommandError:
            self.update_sensors(manager, None)
            raise

        self.update_sensors(manager, output)

        return output



from __future__ import annotations

import slugify

from .errors import NameKeyError


def name_to_key(name: str) -> str:
    """Name to key."""
    if not name:
        raise NameKeyError

    return slugify.slugify(name, separator="_")
